import os
import random
import subprocess

"""
ref : https://github.com/soimort/you-get
"""


class FFmpeg():
    def __init__(self, ffmpeg_path='ffmpeg'):
        assert subprocess.run([ffmpeg_path], stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE).returncode == 1, "请配置正确的 FFmpeg 路径"
        self.path = os.path.normpath(ffmpeg_path)
        tmp_dir = os.path.join(os.path.dirname(ffmpeg_path), "tmp")
        if not os.path.exists(tmp_dir):
            os.makedirs(tmp_dir)
        self.tmp_dir = os.path.normpath(tmp_dir)

    def exec(self, params):
        """ 调用 ffmpeg """
        cmd = [self.path]
        cmd.extend(params)
        return subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    def convert(self, input_path, output_path):
        """ 视频格式转换 """
        params = [
            '-i', input_path,
            '-c', 'copy',
            '-map', '0',
            '-y',
            output_path
        ]
        self.exec(params)

    def join_videos(self, video_path_list, output_path):
        """ 将视频拼接起来 """
        concat_list_path = os.path.join(
            self.tmp_dir, 'concat_list_{:04}.tmp'.format(random.randint(0, 9999))).replace('\\', '/')
        with open(concat_list_path, 'w', encoding='utf-8') as f:
            for video_path in video_path_list:
                if os.path.isfile(video_path):
                    video_relpath = os.path.relpath(
                        video_path, start=self.tmp_dir)
                    f.write("file '{}'\n".format(video_relpath))
        params = [
            '-f', 'concat',
            '-safe', '-1',
            '-i', concat_list_path,
            '-c', 'copy',
            '-y',
            output_path
        ]
        self.exec(params)
        os.remove(concat_list_path)

    def join_video_audio(self, video_path, audio_path, output_path):
        """ 将视频和音频合并 """

        params = [
            '-i', video_path,
            '-i', audio_path,
            '-codec', 'copy',
            '-y',
            output_path
        ]

        self.exec(params)



import re
import os

from bilili.tools import spider
from bilili.utils.quality import quality_map
from bilili.video import BililiContainer
from bilili.utils.base import repair_filename, touch_dir, touch_url

info_api = "https://api.bilibili.com/pgc/web/season/section?season_id={season_id}"
parse_api = "https://api.bilibili.com/pgc/player/web/playurl?avid={avid}&cid={cid}&qn={qn}&ep_id={ep_id}"


def get_title(home_url):
    """ 获取视频标题 """
    res = spider.get(home_url)
    title = re.search(
        r'<span class="media-info-title-t">(.*?)</span>', res.text).group(1)
    return title

def get_context(home_url):
    context = {
        'season_id': '',
    }

    season_id = re.search(
        r'"param":{"season_id":(\d+),"season_type":\d+}', spider.get(home_url).text).group(1)
    context['season_id'] = season_id

    return context

def get_containers(context, video_dir, format, playlist=None):
    season_id = context['season_id']
    containers = []
    info_url = info_api.format(season_id=season_id)
    res = spider.get(info_url)

    for i, item in enumerate(res.json()["result"]["main_section"]["episodes"]):
        index = item["title"]
        if re.match(r'^\d*\.?\d*$', index):
            index = '第{}话'.format(index)
        name = repair_filename(' '.join([index, item["long_title"]]))
        file_path = os.path.join(video_dir, repair_filename(
            '{}.mp4'.format(name)))
        if playlist is not None:
            playlist.write_path(file_path)
        containers.append(BililiContainer(
            id=i+1,
            name=name,
            path=file_path,
            meta={
                "aid": item["aid"],
                "cid": item["cid"],
                "epid": item["id"],
                "bvid": ''
            },
            format=format,
        ))
    if playlist is not None:
        playlist.flush()

    return containers

def parse_segments(container, quality_sequence):
    aid, cid, ep_id, bvid = container.meta["aid"], container.meta["cid"], container.meta["epid"], container.meta["bvid"]

    if container.format == "flv":
        # 检查是否可以下载，同时搜索支持的清晰度，并匹配最佳清晰度
        touch_message = spider.get(parse_api.format(
            avid=aid, cid=cid, ep_id=ep_id, qn=80)).json()
        if touch_message["code"] != 0:
            print("warn: 无法下载 {} ，原因： {}".format(
                container.name, touch_message["message"]))
            return
        if touch_message["result"]["is_preview"] == 1:
            print("warn: {} 为预览版视频".format(container.name))

        accept_quality = touch_message['result']['accept_quality']
        for qn in quality_sequence:
            if qn in accept_quality:
                break

        parse_url = parse_api.format(avid=aid, cid=cid, ep_id=ep_id, qn=qn)
        res = spider.get(parse_url)

        for i, segment in enumerate(res.json()['result']['durl']):
            container.append_media(
                id=i+1,
                url=segment["url"],
                qn=qn,
                height=quality_map[qn]['height'],
                width=quality_map[qn]['width'],
                size=segment["size"],
                type="segment",
            )
    elif container.format == "m4s":
        # 检查是否可以下载，同时搜索支持的清晰度，并匹配最佳清晰度
        parse_api_m4s = parse_api + "&fnver=0&fnval=16&fourk=1"
        play_info = spider.get(parse_api_m4s.format(
            avid=aid, cid=cid, ep_id=ep_id, qn=quality_sequence[0], bvid=bvid)).json()
        if play_info["code"] != 0:
            print("warn: 无法下载 {} ，原因： {}".format(
                container.name, play_info["message"]))
            return
        if play_info["result"]["is_preview"] == 1:
            print("warn: {} 为预览版视频".format(container.name))

        # accept_quality = play_info['result']['accept_quality']
        accept_quality = set([video['id']
                            for video in play_info['result']['dash']['video']])
        for qn in quality_sequence:
            if qn in accept_quality:
                break

        for video in play_info['result']['dash']['video']:
            if video['id'] == qn:
                container.append_media(
                    id=1,
                    url=video['base_url'],
                    qn=qn,
                    height=video['height'],
                    width=video['width'],
                    size=touch_url(video['base_url'], spider)[0],
                    type="video"
                )
                break
        for audio in play_info['result']['dash']['audio']:
            container.append_media(
                id=2,
                url=audio['base_url'],
                height=None,
                width=None,
                size=touch_url(audio['base_url'], spider)[0],
                qn=qn,
                type="audio"
            )
            break

    elif container.format == 'mp4':
        print("番剧不支持 mp4 format")
    else:
        print("Unknown format {}".format(container.format))



import pytest

from bilili.api.acg_video import get_title, get_context, get_containers, parse_segments
from bilili.utils.base import touch_dir
from bilili.utils.quality import quality_sequence_default
from bilili.video import BililiContainer


def test_get_title():
    url = "https://www.bilibili.com/video/BV1Y441167U2/"
    assert get_title(url) == "慕课课程下载工具 Course Crawler 使用方法简介"

def test_get_context():
    url = "https://www.bilibili.com/video/BV1Y441167U2/"
    context = get_context(url)
    assert context["avid"] == ''
    assert context["bvid"] == '1Y441167U2'

def test_get_containers():
    context = {
        "avid": '',
        "bvid": '1Y441167U2',
    }
    video_dir = touch_dir("tmp/MOOC/Videos/")
    containers = get_containers(context, video_dir, 'flv', None)
    assert len(containers) == 3

@pytest.mark.parametrize(
    'format', [
        'flv', 'mp4', 'm4s'
    ])
def test_parse_segments(format):
    container = BililiContainer(
        id=1,
        name="Unknown",
        path=touch_dir("tmp/MOOC/Videos/video1"),
        meta={
            "avid": '',
            "bvid": '1Y441167U2',
            "cid": '113529568',
        },
        format=format,
    )
    parse_segments(container, quality_sequence_default)



import re

from bilili.utils.ffmpeg import FFmpeg
from bilili.utils.crawler import BililiCrawler
from bilili.utils.danmaku import ASS
from bilili.events.middleware import DownloaderMiddleware


ffmpeg = FFmpeg()
spider = BililiCrawler()
ass = ASS()
global_middleware = DownloaderMiddleware()


regex_acg_video_av = re.compile(r"https?://www.bilibili.com/video/av(?P<avid>\d+)")
regex_acg_video_av_short = re.compile(r"https?://b23.tv/av(?P<avid>\d+)")
regex_acg_video_bv = re.compile(r"https?://www.bilibili.com/video/(bv|BV)(?P<bvid>\w+)")
regex_acg_video_bv_short = re.compile(r"https?://b23.tv/(bv|BV)(?P<bvid>\w+)")
regex_bangumi = re.compile(r"https?://www.bilibili.com/bangumi/media/md(\d+)")



import os

from bilili.tools import ffmpeg


noop = lambda *args, **kwargs: None


class MergingFile():
    def __init__(self, format, src_path_list=[], dst_path=''):
        self.format = format
        self.src_path_list = src_path_list
        self.dst_path = dst_path
        self.events = [
            'before_merge', 'merged'
        ]
        for event in self.events:
            setattr(self, event, noop)

    def merge(self):
        self.before_merge(self)
        if self.format == 'mp4':
            with open(self.src_path_list[0].path, 'rb') as fr:
                with open(self.dst_path, 'wb') as fw:
                    fw.write(fr.read())
        elif self.format == 'flv':
            ffmpeg.join_videos(self.src_path_list, self.dst_path)
        elif self.format == 'm4s':
            ffmpeg.join_video_audio(self.src_path_list[0],
                                    self.src_path_list[1], self.dst_path)
        else:
            print('Unknown format {}'.format(self.format))
        for src_path in self.src_path_list:
            os.remove(src_path)
        self.merged(self)

    def on(self, event, **params):
        assert event in self.events
        def on_event(func):
            def new_func(*args, **kwargs):
                return func(*args, **kwargs, **params)
            setattr(self, event, new_func)
        return on_event



import os


class ASS():
    plugin_url = "https://raw.githubusercontent.com/m13253/danmaku2ass/master/danmaku2ass.py"
    plugin_path = "plugins/danmaku2ass.py"
    def __init__(self):
        self.has_plugin = os.path.exists(ASS.plugin_path)

    def initial_plugin(self):
        if self.has_plugin:
            return
        touch_dir(os.path.dirname(ASS.plugin_path))
        touch_file(os.path.join(os.path.dirname(ASS.plugin_path), "__init__.py"))
        print("下载弹幕转换插件中……")
        res = requests.get(plugin_url)
        with open(plugin_path, "w", encoding="utf8") as f:
            f.write(res.text)
        self.has_plugin = True

    def convert_danmaku_from_xml(self, xml_path, height, width):
        self.initial_plugin()
        from plugins.danmaku2ass import Danmaku2ASS
        ass_path = os.path.splitext(xml_path)[0] + '.ass'
        if not os.path.exists(xml_path):
            return
        Danmaku2ASS(
            xml_path, "autodetect", ass_path,
            width, height, reserve_blank=0,
            font_face=_('(FONT) sans-serif')[7:],
            font_size=width/40, text_opacity=0.8, duration_marquee=15.0,
            duration_still=10.0, comment_filter=None, is_reduce_comments=False,
            progress_callback=None)
        os.remove(xml_path)



import os
import requests


noop = lambda *args, **kwargs: None


class RemoteFile():
    """ 远程文件类

    网络 url 与本地文件的绑定，可调用 download 进行下载
    download 支持断点续传
    通过中间件与外部监控程序通讯
    """

    def __init__(self, url, local_path):
        self.url = url
        self.path = local_path
        self.name = os.path.split(self.path)[-1]
        self.tmp_path = self.path + '.dl'
        self.size = self.get_local_size()
        self.events = [
            'before_download', 'before_update',
            'updated', 'downloaded'
        ]
        for event in self.events:
            setattr(self, event, noop)

    def get_local_size(self):
        """ 通过 os.path.getsize 获取本地文件大小 """
        try:
            if os.path.exists(self.tmp_path):
                size = os.path.getsize(self.tmp_path)
            elif os.path.exists(self.path):
                size = os.path.getsize(self.path)
            else:
                size = 0
        except FileNotFoundError:
            size = 0
        return size

    def download(self, spider, stream=True, chunk_size=1024):

        self.before_download(self)
        if not os.path.exists(self.path):
            downloaded = False
            while not downloaded:
                # 设置 headers
                headers = dict(spider.headers)
                headers["Range"] = "bytes={}-".format(self.size)

                try:
                    # 尝试建立连接
                    res = spider.get(self.url, stream=stream, headers=headers, timeout=(5, 10))
                    # 下载到临时路径
                    with open(self.tmp_path, 'ab') as f:
                        if stream:
                            for chunk in res.iter_content(chunk_size=chunk_size):
                                if not chunk:
                                    break
                                self.before_update(self)
                                f.write(chunk)
                                self.size += len(chunk)
                                self.updated(self)
                        else:
                            f.write(res.content)
                    downloaded = True
                except requests.exceptions.RequestException:
                    print(
                        "[warn] file {}, request timeout, trying again...".format(self.name))

            # 从临时文件迁移，并删除临时文件
            if os.path.exists(self.path):
                os.remove(self.path)
            os.rename(self.tmp_path, self.path)

        self.downloaded(self)

    def on(self, event, **params):
        assert event in self.events
        def on_event(func):
            def new_func(*args, **kwargs):
                return func(*args, **kwargs, **params)
            setattr(self, event, new_func)
        return on_event



class Middleware():
    """ 多层次中间件类 """
    def __init__(self, parent=None, children=[]):
        self.parent = None
        self.children = []
        if parent is not None:
            self.set_parent(parent)
        if children:
            self.add_children(children)

    def add_child(self, child):
        self.children.append(child)
        child.parent = self

    def set_parent(self, parent):
        parent.add_child(self)

    def add_children(self, children):
        for child in children:
            self.add_child(child)

    @property
    def is_leaf(self):
        return not self.children

    @property
    def is_root(self):
        return self.parent is None

class DownloaderMiddleware(Middleware):
    """ 下载中间件类

    用于下载器与外部程序之间的通讯
    """

    def __init__(self, parent=None, children=[]):
        super().__init__(parent=parent, children=children)
        self.__total_size = 0
        self.__size = 0
        self.__downloading = False
        self.__downloaded = False
        self.__merging = False
        self.__merged = False

    @property
    def total_size(self):
        if self.is_leaf:
            return self.__total_size
        else:
            return sum([child.total_size for child in self.children])

    @total_size.setter
    def total_size(self, value):
        if self.is_leaf:
            self.__total_size = value
        else:
            print("[warn] 无法设定非叶子结点的 total_size")

    @property
    def size(self):
        if self.is_leaf:
            if self.downloaded:
                return self.total_size
            return self.__size
        else:
            return sum([child.size for child in self.children])

    @size.setter
    def size(self, value):
        if self.is_leaf:
            self.__size = value
        else:
            print("[warn] 无法设定非叶子结点的 size")

    @property
    def downloading(self):
        if self.is_leaf:
            return self.__downloading
        else:
            return any([child.downloading for child in self.children])

    @downloading.setter
    def downloading(self, value):
        if self.is_leaf:
            self.__downloading = value
        else:
            if value:
                print("[warn] 无法设定非叶子结点的 downloading 为 True")
            else:
                for child in self.children:
                    child.downloading = False

    @property
    def downloaded(self):
        if self.is_leaf:
            return self.__downloaded
        else:
            return all([child.downloaded for child in self.children])

    @downloaded.setter
    def downloaded(self, value):
        if self.is_leaf:
            self.__downloaded = value
        else:
            if value:
                for child in self.children:
                    child.downloaded = True
            else:
                print("[warn] 无法设定非叶子结点的 downloaded 为 False")

    @property
    def merging(self):
        if self.is_leaf:
            return self.__merging
        else:
            return any([child.merging for child in self.children])

    @merging.setter
    def merging(self, value):
        if self.is_leaf:
            self.__merging = value
        else:
            if value:
                print("[warn] 无法设定非叶子结点的 merging 为 True")
            else:
                for child in self.children:
                    child.merging = False

    @property
    def merged(self):
        if self.is_leaf:
            return self.__merged
        else:
            return all([child.merged for child in self.children])

    @merged.setter
    def merged(self, value):
        if self.is_leaf:
            self.__merged = value
        else:
            if value:
                for child in self.children:
                    child.merged = True
            else:
                print("[warn] 无法设定非叶子结点的 merged 为 False")



import re
import os

from bilili.tools import (spider, regex_acg_video_av, regex_acg_video_av_short,
                            regex_acg_video_bv, regex_acg_video_bv_short)
from bilili.utils.quality import quality_map
from bilili.video import BililiContainer
from bilili.utils.base import repair_filename, touch_dir, touch_url

info_api = "https://api.bilibili.com/x/player/pagelist?aid={avid}&bvid={bvid}&jsonp=jsonp"
parse_api = "https://api.bilibili.com/x/player/playurl?avid={avid}&cid={cid}&bvid={bvid}&qn={qn}&type=&otype=json"


def get_title(home_url):
    res = spider.get(home_url)
    title = re.search(
        r'<title .*>(.*)_哔哩哔哩 \(゜-゜\)つロ 干杯~-bilibili</title>', res.text).group(1)
    return title

def get_context(home_url):
    context = {
        'avid': '',
        'bvid': ''
    }

    if regex_acg_video_av.match(home_url):
        context['avid'] = regex_acg_video_av.match(home_url).group('avid')
    elif regex_acg_video_av_short.match(home_url):
        context['avid'] = regex_acg_video_av_short.match(home_url).group('avid')
    elif regex_acg_video_bv.match(home_url):
        context['bvid'] = regex_acg_video_bv.match(home_url).group('bvid')
    elif regex_acg_video_bv_short.match(home_url):
        context['bvid'] = regex_acg_video_bv_short.match(home_url).group('bvid')

    return context

def get_containers(context, video_dir, format, playlist=None):
    avid, bvid = context['avid'], context['bvid']
    containers = []
    info_url = info_api.format(avid=avid, bvid=bvid)
    res = spider.get(info_url)

    for i, item in enumerate(res.json()["data"]):
        file_path = os.path.join(video_dir, repair_filename(
            '{}.mp4'.format(item["part"])))
        if playlist is not None:
            playlist.write_path(file_path)
        containers.append(BililiContainer(
            id=i+1,
            name=item["part"],
            path=file_path,
            meta={
                "avid": avid,
                "bvid": bvid,
                "cid": item["cid"]
            },
            format=format,
        ))
    if playlist is not None:
        playlist.flush()
    return containers

def parse_segments(container, quality_sequence):
    cid, avid, bvid = container.meta["cid"], container.meta["avid"], container.meta["bvid"]

    if container.format == "flv":
        # 检查是否可以下载，同时搜索支持的清晰度，并匹配最佳清晰度
        touch_message = spider.get(parse_api.format(
            avid=avid, cid=cid, bvid=bvid, qn=80)).json()
        if touch_message["code"] != 0:
            print("warn: 无法下载 {} ，原因： {}".format(
                container.name, touch_message["message"]))
            return

        accept_quality = touch_message['data']['accept_quality']
        for qn in quality_sequence:
            if qn in accept_quality:
                break

        parse_url = parse_api.format(avid=avid, cid=cid, bvid=bvid, qn=qn)
        res = spider.get(parse_url)

        for i, segment in enumerate(res.json()['data']['durl']):
            container.append_media(
                id=i+1,
                url=segment["url"],
                qn=qn,
                height=quality_map[qn]['height'],
                width=quality_map[qn]['width'],
                size=segment["size"],
                type="segment",
            )
    elif container.format == "m4s":
        # 检查是否可以下载，同时搜索支持的清晰度，并匹配最佳清晰度
        parse_api_m4s = parse_api + "&fnver=0&fnval=16&fourk=1"
        play_info = spider.get(parse_api_m4s.format(
            avid=avid, cid=cid, bvid=bvid, qn=quality_sequence[0])).json()
        if play_info["code"] != 0:
            print("warn: 无法下载 {} ，原因： {}".format(
                container.name, play_info["message"]))
            return

        if play_info['data'].get('dash') is None:
            raise Exception('该视频尚不支持 M4S format 哦~')

        # accept_quality = play_info['data']['accept_quality']
        accept_quality = set([video['id']
                            for video in play_info['data']['dash']['video']])
        for qn in quality_sequence:
            if qn in accept_quality:
                break

        parse_url = parse_api_m4s.format(avid=avid, cid=cid, bvid=bvid, qn=qn)
        play_info = spider.get(parse_url).json()

        for video in play_info['data']['dash']['video']:
            if video['id'] == qn:
                container.append_media(
                    id=1,
                    url=video['base_url'],
                    qn=qn,
                    height=video['height'],
                    width=video['width'],
                    size=touch_url(video['base_url'], spider)[0],
                    type="video"
                )
                break
        for audio in play_info['data']['dash']['audio']:
            container.append_media(
                id=2,
                url=audio['base_url'],
                height=None,
                width=None,
                size=touch_url(audio['base_url'], spider)[0],
                qn=qn,
                type="audio"
            )
            break

    elif container.format == 'mp4':
        # 检查是否可以下载，同时搜索支持的清晰度，并匹配最佳清晰度
        parse_api_mp4 = parse_api + "&platform=html5&high_quality=1"
        play_info = spider.get(parse_api_mp4.format(
            avid=avid, cid=cid, bvid=bvid, qn=120)).json()
        if play_info["code"] != 0:
            print("warn: 无法下载 {} ，原因： {}".format(
                container.name, play_info["message"]))
            return

        container.append_media(
            id=1,
            url=play_info['data']['durl'][0]['url'],
            qn=play_info['data']['quality'],
            height=quality_map[play_info['data']['quality']]['height'],
            width=quality_map[play_info['data']['quality']]['width'],
            size=play_info['data']['durl'][0]['size'],
            type="container"
        )
    else:
        print("Unknown format {}".format(container.format))



import os
import re
import time
import math
import subprocess

from bilili.events.middleware import DownloaderMiddleware
from bilili.utils.quality import quality_map
from bilili.tools import ffmpeg, global_middleware


class BililiContainer():
    """ bilibili 媒体容器类
    即 B 站上的单个视频，其中可能包含多个媒体单元
    * 包含多个 flv 片段
    * 包含 m4s 的视频与音频流
    * 包含完整的一个 mp4
    """

    def __init__(self, id, name, path, meta, format="flv"):

        self.id = id
        self.name = name
        self.path = path
        self.meta = meta
        self.format = format

        self.medias = []
        self.qn = None
        self.height = None
        self.width = None
        self._ = DownloaderMiddleware(parent=global_middleware)

    def append_media(self, *args, **kwargs):
        self.medias.append(BililiMedia(*args, **kwargs, container = self))

    def __str__(self):
        return '{} 「{}」'.format(self.name, quality_map[self.qn]['description'])


class BililiMedia():
    """ bilibili 媒体单元类
    从 B 站直接获取的可下载的媒体单元，可能是 flv、mp4、m4s
    """

    def __init__(self, id, url, qn, size, height, width, container, type="segment"):

        self.id = id
        self.qn = qn
        self.height = height
        self.width = width
        self.url = url
        self.container = container
        self.path = os.path.splitext(self.container.path)[0]
        if self.container.format == "flv":
            self.path += "_{:02d}.flv".format(id)
        elif self.container.format == "m4s":
            self.path += "_{}.m4s".format(type)
        elif self.container.format == "mp4":
            self.path += "_dl.mp4"
        self.name = os.path.split(self.path)[-1]
        self._ = DownloaderMiddleware(parent=self.container._)
        self._.total_size = size

        if self.container.qn is None:
            self.container.qn = qn
        if self.container.width is None:
            self.container.width = width
        if self.container.height is None:
            self.container.height = height
        if self._.total_size is None:
            print("[warn] {} 无法获取 size".format(self.name))
            self._.total_size = 0



import os
import sys
from shutil import rmtree

from bilili.__version__ import __version__
from setuptools import setup, find_packages, Command

here = os.path.abspath(os.path.dirname(__file__))

class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(__version__))
        os.system('git push --tags')

        sys.exit()

def get_long_description():
    with open('README.md', 'r', encoding='utf-8') as f:
        desc = f.read()
    return desc

setup(
    name='bilili',
    version=__version__,
    description="🍻 bilibili video and danmaku downloader | B站视频、弹幕下载器",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    keywords='python bilibili video download spider danmaku',
    author='SigureMo',
    author_email='sigure.qaq@gmail.com',
    url='https://github.com/SigureMo/bilili',
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=True,
    python_requires='>=3.6.0',
    setup_requires=[
        'wheel'
    ],
    install_requires=[
        'requests',
    ],
    entry_points={
        'console_scripts':[
            'bilili = bilili.bilili_dl:main'
        ]
    },
    cmdclass={
        'upload': UploadCommand,
    },
)



import pytest

from bilili.api.danmaku import get_danmaku
from bilili.utils.base import touch_dir
from bilili.video import BililiContainer


def test_danmaku():
    container = BililiContainer(
        id=1,
        name="Unknown",
        path=touch_dir("tmp/ThreeBody/Videos/video1"),
        meta={
            "aid": 84271171,
            "cid": 144541892,
            "epid": 300998,
            "bvid": '',
        },
        format=format,
    )
    get_danmaku(container)



import re
import sys
import argparse
import os
import json
import time
import shutil

from bilili.utils.base import repair_filename, touch_dir, touch_file, size_format
from bilili.utils.quality import quality_sequence_default
from bilili.utils.playlist import Dpl, M3u
from bilili.utils.thread import ThreadPool, Flag
from bilili.utils.console import Console, Font, Line, String, ProgressBar, List, DynamicSymbol, ColorString
from bilili.api.subtitle import get_subtitle
from bilili.api.danmaku import get_danmaku
from bilili.tools import (spider, ass, regex_acg_video_av, regex_acg_video_av_short,
                          regex_acg_video_bv, regex_acg_video_bv_short, regex_bangumi)
from bilili.video import global_middleware
from bilili.events.downloader import RemoteFile
from bilili.events.merger import MergingFile


def parse_episodes(episodes_str, total):
    """ 将选集字符串转为列表 """

    # 解析字符串为列表
    print("全 {} 话".format(total))
    if episodes_str == "all":
        episode_list = list(range(1, total+1))
    elif re.match(r"\d+~\d+", episodes_str):
        start, end = episodes_str.split("~")
        start, end = int(start), int(end)
        assert end > start, "终点值应大于起点值"
        episode_list = list(range(start, end+1))
    elif re.match(r"\d+(,\d+)*", episodes_str):
        episode_list = episodes_str.split(",")
        episode_list = list(map(int, episode_list))
    else:
        episode_list = []

    # 筛选满足条件的剧集
    out_of_range = []
    episodes = []
    for episode in episode_list:
        if episode in range(1, total+1):
            if episode not in episodes:
                episodes.append(episode)
        else:
            out_of_range.append(episode)
    if out_of_range:
        print("warn: 剧集 {} 不存在".format(",".join(list(map(str, out_of_range)))))

    print("已选择第 {} 话".format(",".join(list(map(str, episodes)))))
    assert episodes, "没有选中任何剧集"
    return episodes


def main():
    """ 解析命令行参数并调用相关模块进行下载 """

    parser = argparse.ArgumentParser(description="bilili B 站视频、弹幕下载器")
    parser.add_argument("url", help="视频主页地址")
    parser.add_argument('-f', '--format', default='m4s',
                        choices=['flv', 'm4s', 'mp4'], help="选择下载源格式（m4s 或 flv 或 mp4）")
    parser.add_argument("-d", "--dir", default=r"", help="下载目录")
    parser.add_argument("-q", "--quality", default='120', choices=['120', '116', '112', '80', '74', '64', '32', '16', '6'],
                        help="视频清晰度 120:4K, 116:1080P60, 112:1080P+, 80:1080P, 74:720P60, 64:720P, 32:480P, 16:360P, 6:240P")
    parser.add_argument("-t", "--num-threads", default=30,
                        type=int, help="最大下载线程数")
    parser.add_argument("-p", "--episodes", default="all", help="选集")
    parser.add_argument("-w", "--overwrite",
                        action="store_true", help="强制覆盖已下载视频")
    parser.add_argument("-c", "--sess-data", default=None, help="输入 cookies")
    parser.add_argument("-y", "--yes", action="store_true", help="跳过下载询问")
    parser.add_argument("--playlist-type", default="dpl",
                        choices=["dpl", "m3u", "no"], help="播放列表类型，支持 dpl 和 m3u，输入 no 不生成播放列表")
    parser.add_argument("--path-type", default="rp",
                        help="播放列表路径类型（rp：相对路径，ap：绝对路径）")
    parser.add_argument("--danmaku", default="xml",
                        choices=["xml", "ass", "no"], help="弹幕类型，支持 xml 和 ass，如果设置为 no 则不下载弹幕")
    parser.add_argument("--debug", action="store_true", help="debug 模式")

    args = parser.parse_args()
    cookies = {
        "SESSDATA": args.sess_data
    }

    config = {
        "url": args.url,
        "dir": args.dir,
        "quality_sequence": quality_sequence_default[quality_sequence_default.index(int(args.quality)):] +
        list(reversed(quality_sequence_default[:quality_sequence_default.index(int(args.quality))])),
        "episodes": args.episodes,
        "playlist_type": args.playlist_type,
        "playlist_path_type": args.path_type.upper(),
        "overwrite": args.overwrite,
        "cookies": cookies,
        "format": args.format.lower(),
    }

    if regex_acg_video_av.match(args.url) or \
       regex_acg_video_av_short.match(args.url) or \
       regex_acg_video_bv.match(args.url) or \
       regex_acg_video_bv_short.match(args.url):
        bili_type = "acg_video"
        from bilili.api.acg_video import get_title, get_context, get_containers, parse_segments
    elif regex_bangumi.match(args.url):
        bili_type = "bangumi"
        from bilili.api.bangumi import get_title, get_context, get_containers, parse_segments
    else:
        print("视频地址有误！")
        sys.exit(1)

    home_url = args.url

    # 获取标题
    spider.set_cookies(config["cookies"])
    title = get_title(home_url)
    print(title)

    # 创建所需目录结构
    base_dir = touch_dir(os.path.join(
        config['dir'], repair_filename(title + " - bilibili")))
    video_dir = touch_dir(os.path.join(base_dir, "Videos"))
    if args.overwrite:
        shutil.rmtree(video_dir)
        touch_dir(video_dir)
    if config['playlist_type'] == 'dpl':
        playlist = Dpl(os.path.join(base_dir, 'Playlist.dpl'),
                       path_type=config["playlist_path_type"])
    elif config["playlist_type"] == "m3u":
        playlist = M3u(os.path.join(base_dir, 'Playlist.m3u'),
                       path_type=config["playlist_path_type"])
    else:
        playlist = None

    # 获取需要的信息
    context = get_context(home_url)
    containers = get_containers(context, video_dir, config['format'], playlist)

    # 解析并过滤不需要的选集
    episodes = parse_episodes(config["episodes"], len(containers))
    containers, containers_need_filter = [], containers
    for container in containers_need_filter:
        if container.id not in episodes:
            container._.downloaded = True
            container._.merged = True
        else:
            containers.append(container)

    # 解析片段信息及视频 url
    for i, container in enumerate(containers):
        print("{:02}/{:02} parsing segments info...".format(i,
                                                            len(containers)), end="\r")
        if bili_type == 'acg_video':
            get_subtitle(container)
        if args.danmaku != 'no':
            get_danmaku(container)

        parse_segments(container, config['quality_sequence'])

        if args.danmaku == 'ass':
            ass.convert_danmaku_from_xml(
                os.path.splitext(container.path)[0]+'.xml', container.height, container.width)

    # 准备下载
    if containers:
        # 状态检查与校正
        for i, container in enumerate(containers):
            container_downloaded = os.path.exists(container.path)
            sign = "✓" if container_downloaded else "✖"
            if container_downloaded:
                container._.merged = True
            print("{} {}".format(sign, str(container)))
            for media in container.medias:
                media_downloaded = os.path.exists(media.path)
                sign = "✓" if media_downloaded else "✖"
                media._.downloaded = media_downloaded or container_downloaded
                if not container_downloaded:
                    print("    {} {}".format(sign, media.name))

        # 询问是否下载，通过参数 -y 可以跳过
        if not args.yes:
            answer = None
            while answer is None:
                result = input("以上标 ✖ 为需要进行下载的视频，是否立刻进行下载？[Y/n]")
                if result == '' or result[0].lower() == 'y':
                    answer = True
                elif result[0].lower() == 'n':
                    answer = False
                else:
                    answer = None
            if not answer:
                sys.exit(0)

        # 部署下载与合并任务
        merge_wait_flag = Flag(False)                       # 合并线程池不能因为没有任务就结束
        merge_pool = ThreadPool(3, wait=merge_wait_flag)    # 因此要设定一个 flag，待最后合并结束后改变其值
        download_pool = ThreadPool(args.num_threads)
        for container in containers:
            merging_file = MergingFile(container.format,
                            [media.path for media in container.medias], container.path)
            for media in container.medias:
                remote_file = RemoteFile(media.url, media.path)

                # 为下载挂载各种钩子，以修改状态
                @remote_file.on('before_download')
                def before_download(file, middleware=media._):
                    middleware.downloading = True

                @remote_file.on('updated', middleware=media._)
                def updated(file, middleware=None):
                    middleware.size = file.size

                @remote_file.on('downloaded', middleware=media._, merging_file=merging_file)
                def downloaded(file, middleware=None, merging_file=None):
                    middleware.downloaded = True
                    middleware.downloading = False

                    # 下载完的，部署合并任务
                    if middleware.parent.downloaded and not middleware.parent.merged:

                        # 为合并挂载各种钩子
                        @merging_file.on('before_merge', middleware=middleware.parent)
                        def before_merge(file, middleware=None):
                            for child in middleware.children:
                                child.merging = True

                        @merging_file.on('merged', middleware=middleware.parent)
                        def merged(file, middleware=None):
                            middleware.merging = False
                            middleware.merged = True

                        merge_pool.add_task(merging_file.merge, args=())

                # 下载过的不应继续部署任务
                if media._.downloaded:
                    continue
                download_pool.add_task(remote_file.download, args=(spider, ))

        # 启动线程池
        merge_pool.run()
        download_pool.run()

        # 初始化界面
        console = Console(debug=args.debug)
        console.add_component(
            Line(center=Font(char_a='𝓪', char_A='𝓐'), fillchar=' '))
        console.add_component(Line(left=ColorString(fore='cyan'), fillchar=' '))
        console.add_component(
            List(Line(left=String(), right=String(), fillchar='-')))
        console.add_component(Line(left=ColorString(fore='green', back='white', subcomponent=ProgressBar(
            symbols=' ▏▎▍▌▋▊▉█', width=65)), right=String(), fillchar=' '))
        console.add_component(Line(left=ColorString(fore='blue'), fillchar=' '))
        console.add_component(
            List(Line(left=String(), right=DynamicSymbol(), fillchar=' ')))
        console.add_component(Line(left=ColorString(fore='green', back='white', subcomponent=ProgressBar(
            symbols=' ▏▎▍▌▋▊▉█', width=65)), right=String(), fillchar=' '))

        # 准备监控
        size, t = global_middleware.size, time.time()
        while True:
            now_size, now_t = global_middleware.size, time.time()
            delta_size, delta_t = max(
                now_size - size, 0), (now_t - t) if now_t - t > 1e-6 else 1e-6
            speed = delta_size / delta_t
            size, t = now_size, now_t

            # 数据传入，界面渲染
            console.refresh([
                {
                    'center': ' 🍻 bilili ',
                },
                {
                    'left': '🌠 Downloading videos: '
                } if global_middleware.downloading else None,
                [
                    {
                        'left': '{} '.format(str(container)),
                        'right': ' {}/{}'.format(
                            size_format(container._.size),
                            size_format(container._.total_size)
                        )
                    } if container._.downloading else None for container in containers
                ] if global_middleware.downloading else None,
                {
                    'left': global_middleware.size / global_middleware.total_size,
                    'right': " {}/{} {}/s ⚡".format(
                        size_format(global_middleware.size),
                        size_format(global_middleware.total_size),
                        size_format(speed)
                    )
                } if global_middleware.downloading else None,
                {
                    'left': '🍰 Merging videos: '
                } if global_middleware.merging else None,
                [
                    {
                        'left': '{} '.format(str(container)),
                        'right': True
                    } if container._.merging else None for container in containers
                ] if global_middleware.merging else None,
                {
                    'left': sum([container._.merged for container in containers]) / len(containers),
                    'right': " {}/{} 🚀".format(
                        sum([container._.merged for container in containers]),
                        len(containers)
                    )
                } if global_middleware.merging else None,
            ])

            # 检查是否已经全部完成
            if global_middleware.downloaded and global_middleware.merged:
                merge_wait_flag.value = True
                break
            try:
                # 将刷新率稳定在 2fps
                refresh_rate = 2
                time.sleep(max(1 / refresh_rate - (time.time()-now_t), 0.01))
            except (SystemExit, KeyboardInterrupt):
                raise
        print("已全部下载完成！")
    else:
        print("没有需要下载的视频！")


if __name__ == "__main__":
    main()



import os

from bilili.tools import spider

danmaku_api = "http://comment.bilibili.com/{cid}.xml"

def get_danmaku(container):
    # 下载弹幕
    danmaku_url = danmaku_api.format(cid=container.meta['cid'])
    res = spider.get(danmaku_url)
    res.encoding = "utf-8"
    danmaku_path = os.path.splitext(container.path)[0] + ".xml"
    with open(danmaku_path, "w", encoding="utf-8") as f:
        f.write(res.text)
